
#include <AccelStepper.h>

#define STEPS_PER__ROTATION 2048
#define HALF_STEP 8

#define PIN__BUTTON__OPEN   2

#define PIN__MOTOR__IN1 4
#define PIN__MOTOR__IN2 5
#define PIN__MOTOR__IN3 6
#define PIN__MOTOR__IN4 7

#define DELAY__VEND__EXTEND  1000
#define DELAY__VEND__RETRACT 1000

#define STEPS__VEND__EXTEND 1000
#define STEPS__VEND__RETRACT -1000


AccelStepper stepper = AccelStepper(
  HALF_STEP,
  PIN__MOTOR__IN1,
  PIN__MOTOR__IN3,
  PIN__MOTOR__IN2,
  PIN__MOTOR__IN4);

int uart_steps = 0;
int loop_steps = 0;
volatile bool is_vending = false;
bool is_uart__stepping = false;
bool vending__parity = false;

typedef enum Stepper_State {
  Stepper_State__None = 0,
  Stepper_State__Extending,
  Stepper_State__Unknown
} Stepper_State;

Stepper_State vending__state_of__stepper = Stepper_State__None;

void ISR_vend(void) {
  is_vending = true;
}

void get_steps(void) {
  uart_steps = 0;
  bool sign = false;
  bool is_loop = false;
  while (Serial.available()) {
    unsigned char uart_byte = Serial.read();
    switch (uart_byte) {
      case '*':
        is_loop = true;
        break;
      case '-':
        if (sign) // have to flip bools like this for some reason
          sign = 0;
        else
          sign = 1;
        break;
      default:
        if (uart_byte < '0' || uart_byte > '9')
          break;
        if (uart_steps)
          uart_steps *= 10;
        uart_steps += uart_byte - '0';
        Serial.print(uart_byte - '0');
        break;
    }
  }
  if (sign) {
    uart_steps *= -1;
  }
  if (is_loop) {
    loop_steps = uart_steps;
    uart_steps = 0;
    Serial.print("Looping: ");
    Serial.println(loop_steps);
  } else {
    Serial.print("Steps: ");
    Serial.println(uart_steps);
  }
}

void poll_uart(void) {
  if (is_uart__stepping) {
    if (stepper.distanceToGo() == 0) {
      is_uart__stepping = 0;
      return;
    }
    stepper.run();
    return;
  }
  if (!Serial.available())
    return;
  is_uart__stepping = true;
  uart_steps = 0;
  delay(100); // need to delay for UART
  get_steps();
  stepper.moveTo(uart_steps);
  if (uart_steps < 0)
    uart_steps *= -1;
  Serial.println("Done");
}

void poll_vend(void) {
  if (!is_vending) {
    return;
  }

  switch (vending__state_of__stepper) {
    default:
      vending__state_of__stepper = Stepper_State__None;
    case Stepper_State__None:
      Serial.println("Vending!");
      stepper.setCurrentPosition(
        (vending__parity)
        ? -STEPS__VEND__EXTEND
        : STEPS__VEND__EXTEND);
      vending__state_of__stepper = Stepper_State__Extending;
      stepper.moveTo(
        (vending__parity)
        ? STEPS__VEND__EXTEND
        : -STEPS__VEND__EXTEND);
    case Stepper_State__Extending:
      if (stepper.distanceToGo() != 0) {
        stepper.run();
        break;
      }
      vending__state_of__stepper = Stepper_State__None;
      Serial.println("Done Vending!");
      if (loop_steps) {
        Serial.print("(");
        Serial.print(loop_steps);
        Serial.println(")");
      }
      is_vending = false;
      // for whatever reason we have to flip parity in this fashion:
      if (vending__parity)
        vending__parity = 0;
      else
        vending__parity = 1;
      break;
  }
}

void setup(void) {
  pinMode(
    PIN__BUTTON__OPEN,
    INPUT);
    attachInterrupt(
      digitalPinToInterrupt(
        PIN__BUTTON__OPEN),
      ISR_vend,
      RISING);

  stepper.setMaxSpeed(12800.0 * 32);
  stepper.setAcceleration(800);
  stepper.setSpeed(12800 * 32);
  
  Serial.begin(9600);
  Serial.println("Ready");
}

void poll_loop(void) {
  if (!loop_steps) {
    return;  
  }
  if (is_vending) {
    return;
  }
  delay(2000);
  loop_steps--;
  is_vending = true;
}

void loop(void) {
  poll_uart();
  poll_vend();
  poll_loop();
}
